from typing import Union

import numpy as np
from numba import njit
from jesse.helpers import np_shift, slice_candles

@njit(cache=True)
def wilder_smoothing_numba(raw: np.ndarray, period: int) -> np.ndarray:
    smoothed = np.zeros_like(raw)
    alpha = 1 - 1/period
    smoothed[0] = raw[0]
    for i in range(1, len(raw)):
        smoothed[i] = alpha * smoothed[i-1] + raw[i]
    return smoothed

@njit(cache=True)
def rolling_max_numba(arr: np.ndarray, window: int) -> np.ndarray:
    n = len(arr)
    result = np.empty_like(arr)
    
    # Handle first window-1 elements
    for i in range(window-1):
        result[i] = np.max(arr[:i+1])
    
    # Handle remaining elements with sliding window
    for i in range(window-1, n):
        result[i] = np.max(arr[i-window+1:i+1])
    
    return result

@njit(cache=True)
def rolling_min_numba(arr: np.ndarray, window: int) -> np.ndarray:
    n = len(arr)
    result = np.empty_like(arr)
    
    # Handle first window-1 elements
    for i in range(window-1):
        result[i] = np.min(arr[:i+1])
    
    # Handle remaining elements with sliding window
    for i in range(window-1, n):
        result[i] = np.min(arr[i-window+1:i+1])
    
    return result

def safezonestop(candles: np.ndarray, period: int = 22, mult: float = 2.5, max_lookback: int = 3,
                 direction: str = "long", sequential: bool = False) -> Union[float, np.ndarray]:
    """
    Safezone Stops - Numba optimized version

    :param candles: np.ndarray
    :param period: int - default: 22
    :param mult: float - default: 2.5
    :param max_lookback: int - default: 3
    :param direction: str - default: long
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    high = candles[:, 3]
    low = candles[:, 4]

    last_high = np_shift(high, 1, fill_value=np.nan)
    last_low = np_shift(low, 1, fill_value=np.nan)

    diff_high = high - last_high
    diff_low = last_low - low
    diff_high = np.where(np.isnan(diff_high), 0, diff_high)
    diff_low = np.where(np.isnan(diff_low), 0, diff_low)

    raw_plus_dm = np.where((diff_high > diff_low) & (diff_high > 0), diff_high, 0)
    raw_minus_dm = np.where((diff_low > diff_high) & (diff_low > 0), diff_low, 0)

    plus_dm = wilder_smoothing_numba(raw_plus_dm, period)
    minus_dm = wilder_smoothing_numba(raw_minus_dm, period)

    if direction == "long":
        intermediate = last_low - mult * minus_dm
        res = rolling_max_numba(intermediate, max_lookback)
    else:
        intermediate = last_high + mult * plus_dm
        res = rolling_min_numba(intermediate, max_lookback)

    return res if sequential else res[-1]



from typing import Union

import numpy as np

from jesse.helpers import get_candle_source, slice_candles
from jesse.indicators.ma import ma


def apo(candles: np.ndarray, fast_period: int = 12, slow_period: int = 26, matype: int = 0, source_type: str = "close",
        sequential: bool = False) -> Union[float, np.ndarray]:
    """
    APO - Absolute Price Oscillator

    :param candles: np.ndarray
    :param fast_period: int - default: 12
    :param slow_period: int - default: 26
    :param matype: int - default: 0
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    source = get_candle_source(candles, source_type=source_type)

    if matype == 24 or matype == 29:
        res = ma(candles, period=fast_period, matype=matype, source_type=source_type, sequential=True) - ma(candles, period=slow_period, matype=matype, source_type=source_type, sequential=True)
    else:
        res = ma(source, period=fast_period, matype=matype, sequential=True) - ma(source, period=slow_period, matype=matype, sequential=True)

    return res if sequential else res[-1]



from collections import namedtuple

import numpy as np

from jesse.helpers import get_candle_source

RSMK = namedtuple('RSMK', ['indicator', 'signal'])


def rsmk(candles: np.ndarray, candles_compare: np.ndarray, lookback: int = 90, period: int = 3, signal_period: int = 20,
        source_type: str = "close", sequential: bool = False) -> RSMK:
    """
    RSMK - Relative Strength

    :param candles: np.ndarray
    :param candles_compare: np.ndarray
    :param lookback: int - default: 90
    :param period: int - default: 3
    :param signal_period: int - default: 20
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    if not sequential and candles.shape[0] > 240:
        candles = candles[-240:]
        candles_compare = candles_compare[-240:]

    source = get_candle_source(candles, source_type=source_type)
    source_compare = get_candle_source(candles_compare, source_type=source_type)

    # Calculate log ratio of asset to index
    log_ratio = np.log(source / source_compare)

    # Calculate momentum: difference between current log ratio and the one from 'lookback' periods ago using vectorized operation
    mom = np.full_like(log_ratio, np.nan)
    if len(log_ratio) > lookback:
        mom[lookback:] = log_ratio[lookback:] - log_ratio[:-lookback]

    def ema(series: np.ndarray, period_val: float) -> np.ndarray:
        alpha = 2.0 / (max(1.0, period_val) + 1.0)
        n = series.shape[0]
        out = np.full_like(series, np.nan)

        # Identify the first valid (non-NaN) index
        valid_indices = np.flatnonzero(~np.isnan(series))
        if valid_indices.size == 0:
            return out

        start_idx = valid_indices[0]
        segment = series[start_idx:]
        m = segment.shape[0]

        # Create matrices for vectorized computation
        t = np.arange(m).reshape(-1, 1)  # shape (m, 1)
        i = np.arange(m).reshape(1, -1)  # shape (1, m)
        lag = t - i

        # Compute weights:
        # For t >= i, if i == 0 then weight = (1 - alpha)^(t - i), else weight = alpha * (1 - alpha)^(t - i)
        weights = np.where(lag < 0, 0, np.where(i == 0, (1 - alpha)**lag, alpha * (1 - alpha)**lag))

        ema_segment = np.sum(weights * segment.reshape(1, -1), axis=1)
        out[start_idx:] = ema_segment
        return out

    # Calculate RSMK indicator: EMA of the momentum scaled by 100
    rsmk_indicator = ema(mom, period) * 100.0

    # Calculate signal line as EMA of the RSMK indicator
    rsmk_signal = ema(rsmk_indicator, signal_period)

    if not sequential:
        rsmk_indicator = rsmk_indicator[-1]
        rsmk_signal = rsmk_signal[-1]

    return RSMK(rsmk_indicator, rsmk_signal)



from typing import Union

import numpy as np

from jesse.helpers import slice_candles
from jesse.indicators.ma import ma


def kaufmanstop(candles: np.ndarray, period: int = 22, mult: float = 2, direction: str = "long", matype: int = 0,
                sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    Perry Kaufman's Stops

    :param candles: np.ndarray
    :param period: int - default: 22
    :param mult: float - default: 2
    :param direction: str - default: long
    :param matype: int - default: 0
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    if matype == 24 or matype == 29:
        raise ValueError("VWMA (matype 24) and VWAP (matype 29) cannot be used in kaufmanstop indicator.")

    candles = slice_candles(candles, sequential)

    high = candles[:, 3]
    low = candles[:, 4]
    
    hl_diff = ma(high - low, period=period, matype=matype, sequential=True)

    res = low - hl_diff * mult if direction == "long" else high + hl_diff * mult 
    return res if sequential else res[-1]



from typing import Optional
from fastapi import APIRouter, Header
from fastapi.responses import JSONResponse

from jesse.services import auth as authenticator
from jesse.services.multiprocessing import process_manager
from jesse.services.web import LiveRequestJson, LiveCancelRequestJson, GetLogsRequestJson, GetOrdersRequestJson
import jesse.helpers as jh

router = APIRouter(prefix="/live", tags=["Live Trading"])


@router.post("")
def live(request_json: LiveRequestJson, authorization: Optional[str] = Header(None)) -> JSONResponse:
    """
    Start live trading
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    jh.validate_cwd()

    trading_mode = 'livetrade' if request_json.paper_mode is False else 'papertrade'

    # execute live session
    from jesse_live import live_mode
    process_manager.add_task(
        live_mode.run,
        request_json.id,
        request_json.debug_mode,
        request_json.exchange,
        request_json.exchange_api_key_id,
        request_json.notification_api_key_id,
        request_json.config,
        request_json.routes,
        request_json.data_routes,
        trading_mode,
    )

    mode = 'live' if request_json.paper_mode is False else 'paper'
    return JSONResponse({'message': f"Started {mode} trading..."}, status_code=202)


@router.post("/cancel")
def cancel_live(request_json: LiveCancelRequestJson, authorization: Optional[str] = Header(None)):
    """
    Cancel live trading
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    process_manager.cancel_process(request_json.id)

    return JSONResponse({'message': f'Live process with ID of {request_json.id} terminated.'}, status_code=200)


@router.post('/logs')
def get_logs(json_request: GetLogsRequestJson, authorization: Optional[str] = Header(None)) -> JSONResponse:
    """
    Get logs for a live trading session
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse_live.services.data_provider import get_logs as gl

    arr = gl(json_request.id, json_request.type, json_request.start_time)

    return JSONResponse({
        'id': json_request.id,
        'data': arr
    }, status_code=200)


@router.post('/orders')
def get_orders(json_request: GetOrdersRequestJson, authorization: Optional[str] = Header(None)) -> JSONResponse:
    """
    Get orders for a live trading session
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse_live.services.data_provider import get_orders as go

    arr = go(json_request.session_id)

    return JSONResponse({
        'id': json_request.id,
        'data': arr
    }, status_code=200)



from typing import Optional
from fastapi import APIRouter, Header
from fastapi.responses import JSONResponse

from jesse.services import auth as authenticator
from jesse.services.web import ConfigRequestJson
import jesse.helpers as jh

router = APIRouter(prefix="/config", tags=["Configuration"])


@router.post("/get")
def get_config(json_request: ConfigRequestJson, authorization: Optional[str] = Header(None)):
    """
    Get the current configuration
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.modes.data_provider import get_config as gc

    return JSONResponse({
        'data': gc(json_request.current_config, has_live=jh.has_live_trade_plugin())
    }, status_code=200)


@router.post("/update")
def update_config(json_request: ConfigRequestJson, authorization: Optional[str] = Header(None)):
    """
    Update the configuration
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.modes.data_provider import update_config as uc

    uc(json_request.current_config)

    return JSONResponse({'message': 'Updated configurations successfully'}, status_code=200)



from jesse.strategies import Strategy
import jesse.helpers as jh
from jesse import utils


class TestStopOrderShouldConsiderExecutedTakeProfitOrdersInSpot(Strategy):
    def should_long(self):
        return self.price == 10

    def go_long(self):
        self.buy = 2, self.price

    def on_open_position(self, order) -> None:
        self.take_profit = [
            (1, 15),
            (1, 20),
        ]

    def on_reduced_position(self, order) -> None:
        assert self.position.qty == 1
        self.stop_loss = 1, self.price - 1

    def on_close_position(self, order) -> None:
        last_trade = self.trades[-1]

        if self.trades_count == 1:
            assert last_trade.exit_price == 17.5
            assert last_trade.entry_price == 10
            assert last_trade.qty == 2
            assert last_trade.type == "long"

    def should_cancel_entry(self):
        return False



from collections import namedtuple

import numpy as np
from numba import njit

from jesse.helpers import get_candle_source, slice_candles

DamianiVolatmeter = namedtuple('DamianiVolatmeter', ['vol', 'anti'])


def atr(high: np.ndarray, low: np.ndarray, close: np.ndarray, timeperiod: int) -> np.ndarray:
    tr = np.empty_like(high, dtype=float)
    tr[0] = high[0] - low[0]
    if high.shape[0] > 1:
        diff1 = high[1:] - low[1:]
        diff2 = np.abs(high[1:] - close[:-1])
        diff3 = np.abs(low[1:] - close[:-1])
        tr[1:] = np.maximum(diff1, np.maximum(diff2, diff3))
    atr_array = np.full(high.shape, np.nan, dtype=float)
    if high.shape[0] < timeperiod:
        atr_array[-1] = np.mean(tr)
        return atr_array
    n = high.shape[0]
    m = n - timeperiod + 1
    alpha = 1.0 / timeperiod
    initial = np.mean(tr[:timeperiod])
    ema_vector = np.empty(m)
    ema_vector[0] = initial
    if m > 1:
        k = np.arange(m)  # k = 0,..., m-1
        initial_contrib = initial * (1 - alpha) ** k
        # Build a lower-triangular matrix for weights for indices 1 to m-1
        exp_matrix = np.tril((1 - alpha) ** (np.subtract.outer(np.arange(m - 1), np.arange(m - 1))))
        # tr[timeperiod:] has length m-1
        sum_vals = alpha * (exp_matrix @ tr[timeperiod:])
        ema_vector[1:] = initial_contrib[1:] + sum_vals
    atr_array[:timeperiod - 1] = np.nan
    atr_array[timeperiod - 1:] = ema_vector
    return atr_array

def damiani_volatmeter(candles: np.ndarray, vis_atr: int = 13, vis_std: int = 20, sed_atr: int = 40, sed_std: int = 100,
                       threshold: float = 1.4, source_type: str = "close",
                       sequential: bool = False) -> DamianiVolatmeter:
    """
    Damiani Volatmeter

    :param candles: np.ndarray
    :param vis_atr: int - default: 13
    :param vis_std: int - default: 20
    :param sed_atr: int - default: 40
    :param sed_std: int - default: 100
    :param threshold: float - default: 1.4
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    atrvis = atr(candles[:, 3], candles[:, 4], candles[:, 2], vis_atr)
    atrsed = atr(candles[:, 3], candles[:, 4], candles[:, 2], sed_atr)

    vol, t = damiani_volatmeter_fast(source, sed_std, atrvis, atrsed, vis_std, threshold)

    if sequential:
        return DamianiVolatmeter(vol, t)
    else:
        return DamianiVolatmeter(vol[-1], t[-1])

def damiani_volatmeter_fast(source, sed_std, atrvis, atrsed, vis_std, threshold):
    from scipy.signal import lfilter
    from numpy.lib.stride_tricks import sliding_window_view

    lag_s = 0.5
    n = source.shape[0]

    # Compute vol using a linear filter to solve the recurrence:
    # vol[i] - lag_s*vol[i-1] + lag_s*vol[i-3] = atrvis[i]/atrsed[i]
    u = np.zeros(n)
    u[sed_std:] = atrvis[sed_std:] / atrsed[sed_std:]
    b = [1.0]
    a = [1.0, -lag_s, 0.0, lag_s]
    vol = lfilter(b, a, u)

    # Compute t vectorized by calculating moving standard deviations without loops
    t = np.zeros(n)
    if n >= sed_std:
        std_vis = np.std(sliding_window_view(source, vis_std), axis=-1)
        std_sed = np.std(sliding_window_view(source, sed_std), axis=-1)
        idx = np.arange(sed_std, n)
        t[idx] = threshold - (std_vis[idx - vis_std] / std_sed[idx - sed_std])
    return vol, t



from collections import namedtuple
import numpy as np
from jesse.helpers import get_candle_source, slice_candles
from jesse_rust import macd as macd_rust

MACD = namedtuple('MACD', ['macd', 'signal', 'hist'])


def macd(candles: np.ndarray, fast_period: int = 12, slow_period: int = 26, signal_period: int = 9,
         source_type: str = "close",
         sequential: bool = False) -> MACD:
    """
    MACD - Moving Average Convergence/Divergence

    :param candles: np.ndarray
    :param fast_period: int - default: 12
    :param slow_period: int - default: 26
    :param signal_period: int - default: 9
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: MACD(macd, signal, hist)
    """

    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    if source.size == 0:
        return MACD(np.nan, np.nan, np.nan) if not sequential else MACD(np.array([]), np.array([]), np.array([]))

    macd_line, signal_line, hist = macd_rust(source, fast_period, slow_period, signal_period)

    if sequential:
        return MACD(macd_line, signal_line, hist)
    else:
        return MACD(macd_line[-1], signal_line[-1], hist[-1])



from typing import Union
import numpy as np
from numba import njit

from jesse.helpers import get_candle_source, slice_candles

@njit(cache=True)
def _compute_cfo(source: np.ndarray, period: int, scalar: float) -> np.ndarray:
    n = source.shape[0]
    res = np.empty(n, dtype=np.float64)
    # fill initial values with nan for indices where a full period is not available
    for i in range(period - 1):
        res[i] = np.nan

    # Precompute constants for x = 0, 1, ..., period-1
    Sx = 0.0
    Sxx = 0.0
    for j in range(period):
        Sx += j
        Sxx += j * j
    denom = period * Sxx - Sx * Sx

    # For each valid window, compute the linear regression and forecast value
    for i in range(period - 1, n):
        sum_y = 0.0
        sum_xy = 0.0
        for j in range(period):
            y_val = source[i - period + 1 + j]
            sum_y += y_val
            sum_xy += y_val * j
        slope = (period * sum_xy - Sx * sum_y) / denom
        intercept = (sum_y - slope * Sx) / period
        reg_val = intercept + slope * (period - 1)
        # Avoid division by zero
        if source[i] != 0.0:
            res[i] = scalar * (source[i] - reg_val) / source[i]
        else:
            res[i] = np.nan
    return res


def cfo(candles: np.ndarray, period: int = 14, scalar: float = 100, source_type: str = "close", sequential: bool = False) -> Union[float, np.ndarray]:
    """
    CFO - Chande Forcast Oscillator

    :param candles: np.ndarray
    :param period: int - default: 14
    :param scalar: float - default: 100
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)
    source = get_candle_source(candles, source_type=source_type)

    res = _compute_cfo(source, period, scalar)
    
    if sequential:
        return res
    else:
        return None if np.isnan(res[-1]) else res[-1]



import requests
import jesse.helpers as jh
from jesse.modes.import_candles_mode.drivers.interface import CandleExchange
from typing import Union
from jesse import exceptions
from .apex_pro_utils import timeframe_to_interval


class ApexProMain(CandleExchange):
    def __init__(self, name: str, rest_endpoint: str) -> None:
        from jesse.modes.import_candles_mode.drivers.Binance.BinanceSpot import BinanceSpot

        super().__init__(name=name, count=200, rate_limit_per_second=10, backup_exchange_class=BinanceSpot)
        self.name = name
        self.endpoint = rest_endpoint

    def get_starting_time(self, symbol: str) -> int:
        dashless_symbol = jh.dashless_symbol(symbol)
        payload = {
            'symbol': dashless_symbol,
            'interval': 'W',
            'limit': 200,
            'start': 1514811660
        }

        response = requests.get(self.endpoint + '/klines', params=payload)
        self.validate_response(response)

        if 'data' not in response.json():
            raise exceptions.ExchangeInMaintenance(response.json()['msg'])
        elif response.json()['data'] == {}:
            raise exceptions.InvalidSymbol('Exchange does not support the entered symbol. Please enter a valid symbol.')

        data = response.json()['data'][dashless_symbol]
        # Reverse the data list
        data = data[::-1]

        return int(data[1]['t'])

    def fetch(self, symbol: str, start_timestamp: int, timeframe: str = '1m') -> Union[list, None]:
        dashless_symbol = jh.dashless_symbol(symbol)
        interval = timeframe_to_interval(timeframe)

        payload = {
            'symbol': dashless_symbol,
            'interval': interval,
            'start': int(start_timestamp / 1000),
            'limit': self.count
        }

        response = requests.get(self.endpoint + '/klines', params=payload)
        # check data exist in response.json

        if 'data' not in response.json():
            raise exceptions.ExchangeInMaintenance(response.json()['msg'])
        elif response.json()['data'] == {}:
            raise exceptions.InvalidSymbol('Exchange does not support the entered symbol. Please enter a valid symbol.')

        data = response.json()['data'][dashless_symbol]

        return [
            {
                'id': jh.generate_unique_id(),
                'exchange': self.name,
                'symbol': symbol,
                'timeframe': timeframe,
                'timestamp': int(d['t']),
                'open': float(d['o']),
                'close': float(d['c']),
                'high': float(d['h']),
                'low': float(d['l']),
                'volume': float(d['v'])
            } for d in data
        ]

    def get_available_symbols(self) -> list:
        response = requests.get(self.endpoint + '/symbols')
        self.validate_response(response)
        data = response.json()['data']

        # Determine which suffix to filter based on exchange name
        target_suffix = '-USDT' if self.name.startswith('Apex Omni') else '-USDC'

        # For legacy API response format
        if 'usdtConfig' not in data:
            symbols = []
            contracts = data['contractConfig']['perpetualContract']
            for p in contracts:
                symbol = p['symbol']
                if symbol.endswith(target_suffix):
                    symbols.append(symbol)
            return list(sorted(symbols))

        # For new API response format
        pairs = []
        # For Omni (USDT pairs)
        if target_suffix == '-USDT':
            if 'usdtConfig' in data and 'perpetualContract' in data['usdtConfig']:
                contracts = data['usdtConfig']['perpetualContract']
                for p in contracts:
                    symbol = p['symbol']
                    if symbol.endswith(target_suffix):
                        pairs.append(symbol)
        # For Pro (USDC pairs)
        else:
            if 'usdcConfig' in data and 'perpetualContract' in data['usdcConfig']:
                contracts = data['usdcConfig']['perpetualContract']
                for p in contracts:
                    symbol = p['symbol']
                    if symbol.endswith(target_suffix):
                        pairs.append(symbol)

        return list(sorted(pairs))



from collections import namedtuple

import numpy as np

from jesse.helpers import slice_candles

AC = namedtuple('AC', ['osc', 'change'])

def sma(arr: np.ndarray, period: int) -> np.ndarray:
    if len(arr) < period:
        return np.full_like(arr, np.nan, dtype=float)
    conv = np.convolve(arr, np.ones(period, dtype=float)/period, mode='valid')
    return np.concatenate((np.full(period-1, np.nan), conv))

def mom(arr: np.ndarray, period: int = 1) -> np.ndarray:
    return np.concatenate(([np.nan], np.diff(arr)))

def acosc(candles: np.ndarray, sequential: bool = False) -> AC:
    """
    Acceleration / Deceleration Oscillator (AC)

    :param candles: np.ndarray
    :param sequential: bool - default: False

    :return: AC(osc, change)
    """
    candles = slice_candles(candles, sequential)

    med = (candles[:, 3] + candles[:, 4]) / 2
    sma5_med = sma(med, 5)
    sma34_med = sma(med, 34)
    ao = sma5_med - sma34_med
    sma5_ao = sma(ao, 5)
    res = ao - sma5_ao
    mom_value = mom(res, 1)

    if sequential:
        return AC(res, mom_value)
    else:
        return AC(res[-1], mom_value[-1])



from typing import Union

import numpy as np
from numba import njit

from jesse.helpers import slice_candles

@njit(cache=True)
def _compute_aroonosc_nb(high: np.ndarray, low: np.ndarray, period: int) -> np.ndarray:
    n = high.shape[0]
    result = np.empty(n, dtype=np.float64)
    if n < period:
        for i in range(n):
            result[i] = np.nan
        return result
    
    for i in range(period - 1):
        result[i] = np.nan
    
    for i in range(period - 1, n):
        start = i - period + 1
        best_val = high[start]
        best_idx = 0
        worst_val = low[start]
        worst_idx = 0
        for j in range(period):
            cur_high = high[start + j]
            cur_low = low[start + j]
            if cur_high > best_val:
                best_val = cur_high
                best_idx = j
            if cur_low < worst_val:
                worst_val = cur_low
                worst_idx = j
        result[i] = 100.0 * (best_idx - worst_idx) / period
    return result


def aroonosc(candles: np.ndarray, period: int = 14, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    AROONOSC - Aroon Oscillator

    :param candles: np.ndarray
    :param period: int - default: 14
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)
    high = candles[:, 3]
    low = candles[:, 4]
    result = _compute_aroonosc_nb(high, low, period)
    return result if sequential else result[-1]



# silent (pandas) warnings
import warnings
from typing import List, Any, Union, Dict, Optional

import numpy as np
import pandas as pd
from peewee import CharField, FloatField

import jesse.helpers as jh
from jesse.config import config
from jesse.routes import router
from jesse.services import metrics as stats
from jesse.services import selectors
from jesse.services.candle import is_bullish
from jesse.store import store
from jesse.models import Position

warnings.filterwarnings("ignore")


def positions() -> list:
    arr = []

    for r in router.routes:
        if r.strategy is None:
            continue
        p: Position = r.strategy.position
        arr.append({
            'currency': jh.app_currency(),
            'type': p.type,
            'strategy_name': p.strategy.name,
            'symbol': p.symbol,
            'leverage': p.leverage,
            'opened_at': p.opened_at,
            'qty': p.qty,
            'value': p.value,
            'entry': p.entry_price,
            'current_price': p.current_price,
            'liquidation_price': p.liquidation_price,
            'pnl': p.pnl,
            'pnl_perc': p.pnl_percentage
        })

    return arr


def candles() -> dict:
    candles_dict = {}
    candle_keys = []

    # add routes
    for e in router.routes:
        if e.strategy is None:
            return {}

        candle_keys.append({
            'exchange': e.exchange,
            'symbol': e.symbol,
            'timeframe': e.timeframe
        })

    # # add data_routes
    # for e in router.data_candles:
    #     candle_keys.append({
    #         'exchange': e['exchange'],
    #         'symbol': e['symbol'],
    #         'timeframe': e['timeframe']
    #     })

    for k in candle_keys:
        try:
            c = store.candles.get_current_candle(k['exchange'], k['symbol'], k['timeframe'])
            key = jh.key(k['exchange'], k['symbol'], k['timeframe'])
            candles_dict[key] = {
                'time': int(c[0] / 1000),
                'open': c[1],
                'close': c[2],
                'high': c[3],
                'low': c[4],
                'volume': c[5],
            }
        except IndexError:
            return {}
        except Exception:
            raise

    return candles_dict


def livetrade():
    starting_balance = 0
    current_balance = 0
    exchange_name = ''
    leverage = 1
    leverage_type = 'spot'
    available_margin = 0
    for e in store.exchanges.storage:
        starting_balance = round(store.exchanges.storage[e].started_balance, 2)
        current_balance = round(store.exchanges.storage[e].wallet_balance, 2)
        exchange_name = e
        if store.exchanges.storage[e].type == 'futures':
            leverage = store.exchanges.storage[e].futures_leverage
            leverage_type = store.exchanges.storage[e].futures_leverage_mode
            available_margin = round(store.exchanges.storage[e].available_margin, 2)
        # there's only one exchange, so we can break
        break

    # short trades summary
    if len(store.completed_trades.trades):
        df = pd.DataFrame.from_records([t.to_dict for t in store.completed_trades.trades])
        total = len(df)
        winning_trades = len(df.loc[df['PNL'] > 0])
        losing_trades = len(df.loc[df['PNL'] < 0])
        pnl = round(df['PNL'].sum(), 2)
        pnl_perc = round((pnl / starting_balance) * 100, 2)
    else:
        pnl, pnl_perc, total, winning_trades, losing_trades = 0, 0, 0, 0, 0

    routes = [
        {
            'symbol': r.symbol,
            'timeframe': r.timeframe,
            'strategy': r.strategy_name
        } for r in router.routes
    ]

    return {
        'session_id': store.app.session_id,
        'started_at': str(store.app.starting_time),
        'current_time': str(jh.now_to_timestamp()),
        'started_balance': str(starting_balance),
        'current_balance': str(current_balance),
        'debug_mode': str(config['app']['debug_mode']),
        'paper_mode': str(jh.is_paper_trading()),
        'count_error_logs': str(len(store.logs.errors)),
        'count_info_logs': str(len(store.logs.info)),
        'count_active_orders': str(store.orders.count_all_active_orders()),
        'open_positions': str(store.positions.count_open_positions()),
        'pnl': str(pnl),
        'pnl_perc': str(pnl_perc),
        'count_trades': str(total),
        'count_winning_trades': str(winning_trades),
        'count_losing_trades': str(losing_trades),
        'routes': routes,
        'exchange': exchange_name,
        'leverage': leverage,
        "leverage_type": leverage_type,
        'available_margin': available_margin
    }


def portfolio_metrics() -> Union[dict, None]:
    if store.completed_trades.count == 0:
        return None

    return stats.trades(store.completed_trades.trades, store.app.daily_balance)


def trades() -> List[dict]:
    if store.completed_trades.count == 0:
        return []
    return [t.to_dict_with_orders for t in store.completed_trades.trades]


def info() -> List[List[Union[str, Any]]]:
    return [
        [
            jh.timestamp_to_time(w['time'])[11:19],
            f"{w['message'][:70]}.."
            if len(w['message']) > 70
            else w['message'],
        ]
        for w in store.logs.info[::-1][0:5]
    ]


def watch_list() -> List[List[Union[str, str]]]:
    """
    Returns a list of data that are currently being watched in realtime
    only support the first route
    """
    strategy = router.routes[0].strategy

    # return if strategy object is not initialized yet
    if strategy is None:
        return []

    # don't if the strategy hasn't been initiated yet
    if not store.candles.are_all_initiated:
        return []

    try:
        watch_list_array = strategy.watch_list()
    except Exception as e:
        import traceback

        watch_list_array = [
            ('', "The watch list is not available because an error occurred while getting it. Please check your strategy code's watch_list() method."),
            ('', f'ERROR: ```{traceback.format_exc()}```')
        ]

    # loop through the watch list and convert each item into a string
    for index, value in enumerate(watch_list_array):
        # if value is not a tuple with two values in it, raise ValueError
        if not isinstance(value, tuple) or len(value) != 2:
            raise ValueError("watch_list() must return a list of tuples with 2 values in each. Example: [(key1, value1), (key2, value2)]")

        watch_list_array[index] = (str(value[0]), str(value[1]))

    return watch_list_array if len(watch_list_array) else []


def errors() -> List[List[Union[str, Any]]]:
    return [
        [
            jh.timestamp_to_time(w['time'])[11:19],
            f"{w['message'][:70]}.."
            if len(w['message']) > 70
            else w['message'],
        ]
        for w in store.logs.errors[::-1][0:5]
    ]


def orders() -> List[dict]:
    route_orders = []

    for r in router.routes:
        r_orders = store.orders.get_orders(r.exchange, r.symbol)
        for o in r_orders:
            route_orders.append(o.to_dict)

    return route_orders



from typing import Union
import numpy as np
from jesse.helpers import slice_candles
from jesse.indicators import ema


def kvo(candles: np.ndarray, short_period: int = 34, long_period: int = 55, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    KVO - Klinger Volume Oscillator

    :param candles: np.ndarray
    :param short_period: int - default: 34
    :param long_period: int - default: 55
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    # Calculate HLC3
    hlc3 = (candles[:, 3] + candles[:, 4] + candles[:, 2]) / 3

    # Calculate momentum (change in HLC3)
    mom = np.diff(hlc3, prepend=hlc3[0])

    # Calculate trend
    trend = np.zeros_like(mom)
    trend[1:] = np.where(mom[1:] > 0, 1, np.where(mom[1:] < 0, -1, trend[:-1]))

    # Daily Measurement (High - Low)
    dm = candles[:, 3] - candles[:, 4]

    # Cumulative Measurement
    cm = np.zeros_like(dm)
    for i in range(1, len(trend)):
        if trend[i] == trend[i-1]:
            cm[i] = cm[i-1] + dm[i]
        else:
            cm[i] = dm[i] + dm[i-1]

    # Volume Force
    volume = candles[:, 5]
    with np.errstate(divide='ignore', invalid='ignore'):
        expr = np.abs(np.divide(2 * dm, cm, out=np.zeros_like(dm, dtype=float), where=cm != 0) - 1)
    vf = 100 * volume * trend * expr
    vf[cm == 0] = 0

    # Calculate EMAs
    fast_ema = ema(vf, period=short_period, sequential=True)
    slow_ema = ema(vf, period=long_period, sequential=True)
    res = fast_ema - slow_ema
    return res if sequential else res[-1]



from collections import namedtuple

import numpy as np

from jesse.helpers import get_candle_source, slice_candles
from jesse.indicators.ma import ma

MAB = namedtuple('MAB', ['upperband', 'middleband', 'lowerband'])


def mab(candles: np.ndarray, fast_period: int = 10, slow_period: int = 50, devup: float = 1, devdn: float = 1, fast_matype: int = 0, slow_matype: int = 0,
                    source_type: str = "close",
                    sequential: bool = False) -> MAB:
    """
    Moving Average Bands

    :param candles: np.ndarray
    :param fast_period: int - default: 10
    :param slow_period: int - default: 50
    :param devup: float - default: 1
    :param devdn: float - default: 1
    :param fast_matype: int - default: 0
    :param slow_matype: int - default: 0
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: MAB(upperband, middleband, lowerband)
    """
    candles = slice_candles(candles, sequential)

    source = get_candle_source(candles, source_type=source_type)

    if any(matype in (24, 29) for matype in (fast_matype, slow_matype)):
        fastEma = ma(candles, period=fast_period, matype=fast_matype, source_type=source_type, sequential=True)
        slowEma = ma(candles, period=slow_period, matype=slow_matype, source_type=source_type, sequential=True)
    else:
        fastEma = ma(source, period=fast_period, matype=fast_matype, source_type=source_type, sequential=True)
        slowEma = ma(source, period=slow_period, matype=slow_matype, source_type=source_type, sequential=True)
    sqAvg = np.sum(np.power(fastEma - slowEma, 2)[-fast_period:]) / fast_period
    dev = np.sqrt(sqAvg)

    middlebands = fastEma
    upperbands = slowEma + devup * dev
    lowerbands = slowEma - devdn * dev


    if sequential:
        return MAB(upperbands, middlebands, lowerbands)
    else:
        return MAB(upperbands[-1], middlebands[-1], lowerbands[-1])



import numpy as np
import pytest

from jesse.libs import DynamicNumpyArray


def test_append():
    a = DynamicNumpyArray((10, 6))
    a.append(np.array([1, 2, 3, 4, 5, 6]))
    assert a.index == 0
    assert a.array[0][0] == 1
    assert a.array[0][1] == 2
    assert a.array[0][2] == 3
    assert a.array[0][3] == 4
    assert a.array[0][4] == 5
    assert a.array[0][5] == 6

    a.append(np.array([7, 8, 9, 10, 11, 12]))
    assert a.index == 1
    assert a.array[1][0] == 7
    assert a.array[1][1] == 8
    assert a.array[1][2] == 9
    assert a.array[1][3] == 10
    assert a.array[1][4] == 11
    assert a.array[1][5] == 12


def test_flush():
    a = DynamicNumpyArray((10, 6))
    a.append(np.array([1, 2, 3, 4, 5, 6]))
    a.append(np.array([7, 8, 9, 10, 11, 12]))
    assert a.index == 1
    assert a.array[0][0] == 1
    assert a.array[1][0] == 7

    a.flush()

    assert a.index == -1
    assert a.array[0][0] == 0
    assert a.array[1][0] == 0


def test_get_last_item():
    a = DynamicNumpyArray((10, 6))

    with pytest.raises(IndexError):
        a.get_last_item()

    a.append(np.array([1, 2, 3, 4, 5, 6]))
    a.append(np.array([7, 8, 9, 10, 11, 12]))
    assert a.index == 1
    assert a.array[0][0] == 1
    assert a.array[1][0] == 7

    np.testing.assert_array_equal(a.get_last_item(), np.array([7, 8, 9, 10, 11, 12]))


def test_get_past_item():
    a = DynamicNumpyArray((10, 6))

    with pytest.raises(IndexError):
        a.get_past_item(1)

    a.append(np.array([1, 2, 3, 4, 5, 6]))
    a.append(np.array([7, 8, 9, 10, 11, 12]))
    assert a.index == 1
    assert a.array[0][0] == 1
    assert a.array[1][0] == 7

    np.testing.assert_array_equal(a.get_past_item(1), np.array([1, 2, 3, 4, 5, 6]))

    with pytest.raises(IndexError):
        a.get_past_item(2)


def test_get_item():
    a = DynamicNumpyArray((10, 6))

    # test getting while array is empty
    with pytest.raises(IndexError):
        var = a[0]

    a.append(np.array([1, 2, 3, 4, 5, 6]))
    a.append(np.array([7, 8, 9, 10, 11, 12]))
    assert a.index == 1
    assert a.array[0][0] == 1
    assert a.array[1][0] == 7

    np.testing.assert_array_equal(a[0], np.array([1, 2, 3, 4, 5, 6]))
    np.testing.assert_array_equal(a[1], np.array([7, 8, 9, 10, 11, 12]))

    # test with an index that is out of range
    with pytest.raises(IndexError):
        var = a[2]


def test_array_size_increases():
    a = DynamicNumpyArray((3, 6))

    assert a.array.shape == (3, 6)

    a.append(np.array([1, 2, 3, 4, 5, 6]))
    a.append(np.array([7, 8, 9, 10, 11, 12]))
    a.append(np.array([13, 14, 15, 16, 17, 18]))
    assert a.array.shape == (6, 6)
    assert a.index == 2

    a.append(np.array([19, 20, 21, 22, 23, 24]))
    a.append(np.array([25, 26, 27, 28, 29, 30]))
    a.append(np.array([31, 32, 33, 34, 35, 36]))
    assert a.array.shape == (9, 6)
    assert a.index == 5


def test_drop_at():
    a = DynamicNumpyArray((100, 6), drop_at=6)

    # add 5 items
    a.append(np.array([1, 2, 3, 4, 5, 6]))
    a.append(np.array([7, 8, 9, 10, 11, 12]))
    a.append(np.array([13, 14, 15, 16, 17, 18]))
    a.append(np.array([19, 20, 21, 22, 23, 24]))
    a.append(np.array([25, 26, 27, 28, 29, 30]))
    assert a.get_last_item()[0] == 25
    assert a[0][0] == 1

    # add 6th item. when it reaches the drop_at limit, it should drop drop_at/2 items
    a.append(np.array([31, 32, 33, 34, 35, 36]))
    assert a[0][0] == 19




from collections import namedtuple
import numpy as np
from numba import njit
from jesse.helpers import get_candle_source, slice_candles
from jesse.indicators.ma import ma

KeltnerChannel = namedtuple('KeltnerChannel', ['upperband', 'middleband', 'lowerband'])


@njit
def _atr(high: np.ndarray, low: np.ndarray, close: np.ndarray, period: int) -> np.ndarray:
    """
    Calculate ATR using Numba
    """
    n = len(close)
    tr = np.empty(n)
    atr_vals = np.full(n, np.nan)
    
    # Calculate True Range
    tr[0] = high[0] - low[0]
    for i in range(1, n):
        hl = high[i] - low[i]
        hc = abs(high[i] - close[i-1])
        lc = abs(low[i] - close[i-1])
        tr[i] = max(max(hl, hc), lc)
    
    if n < period:
        return atr_vals
        
    # First ATR value is the simple average of the first 'period' true ranges
    atr_vals[period-1] = np.mean(tr[:period])
    
    # Calculate subsequent ATR values using Wilder's smoothing
    for i in range(period, n):
        atr_vals[i] = (atr_vals[i-1] * (period - 1) + tr[i]) / period
        
    return atr_vals


@njit
def _calculate_keltner(source: np.ndarray, high: np.ndarray, low: np.ndarray, close: np.ndarray, 
                      ma_values: np.ndarray, period: int, multiplier: float) -> tuple:
    """
    Core Keltner Channel calculation using Numba
    """
    atr_vals = _atr(high, low, close, period)
    
    up = ma_values + atr_vals * multiplier
    low = ma_values - atr_vals * multiplier
    
    return up, ma_values, low


def keltner(candles: np.ndarray, period: int = 20, multiplier: float = 2, matype: int = 1, 
           source_type: str = "close", sequential: bool = False) -> KeltnerChannel:
    """
    Keltner Channels using Numba for optimization

    :param candles: np.ndarray
    :param period: int - default: 20
    :param multiplier: float - default: 2
    :param matype: int - default: 1
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: KeltnerChannel(upperband, middleband, lowerband)
    """
    candles = slice_candles(candles, sequential)

    source = get_candle_source(candles, source_type=source_type)
    if matype == 24 or matype == 29:
        ma_values = ma(candles, period=period, matype=matype, source_type=source_type, sequential=True)
    else:
        ma_values = ma(source, period=period, matype=matype, sequential=True)
    
    up, mid, low = _calculate_keltner(
        source, 
        candles[:, 3],  # high
        candles[:, 4],  # low
        candles[:, 2],  # close
        ma_values,
        period,
        multiplier
    )

    if sequential:
        return KeltnerChannel(up, mid, low)
    else:
        return KeltnerChannel(up[-1], mid[-1], low[-1])



from typing import Union

import numpy as np
from numba import njit

from jesse.helpers import (get_candle_source, np_shift, same_length,
                           slice_candles)


def maaq(candles: np.ndarray, period: int = 11, fast_period: int = 2, slow_period: int = 30, source_type: str = "close", sequential: bool = False) -> Union[float, np.ndarray]:
    """
    Moving Average Adaptive Q

    :param candles: np.ndarray
    :param period: int - default: 11
    :param fast_period: int - default: 2
    :param slow_period: int - default: 30
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """

    # Accept normal array too.
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    source = source[~np.isnan(source)]

    diff = np.abs(source - np_shift(source, 1, np.nan))
    signal = np.abs(source - np_shift(source, period, np.nan))
    noise = np.concatenate((np.full(period - 1, np.nan, dtype=source.dtype), np.convolve(diff, np.ones(period, dtype=source.dtype), mode='valid')))

    # Safely divide signal by noise
    ratio = np.divide(signal, noise, out=np.zeros_like(signal), where=(noise != 0))

    fastSc = 2 / (fast_period + 1)
    slowSc = 2 / (slow_period + 1)
    temp = np.power((ratio * fastSc) + slowSc, 2)

    res = maaq_fast(source, temp, period)
    res = same_length(candles, res)

    return res if sequential else res[-1]


@njit(cache=True)
def maaq_fast(source, temp, period):
    newseries = np.copy(source)
    for i in range(period, source.shape[0]):
        newseries[i] = newseries[i - 1] + (temp[i] * (source[i] - newseries[i - 1]))
    return newseries



from collections import namedtuple

import numpy as np
from numba import njit

from jesse.helpers import get_candle_source, slice_candles

VossFilter = namedtuple('VossFilter', ['voss', 'filt'])


def voss(candles: np.ndarray, period: int = 20, predict: int = 3, bandwith: float = 0.25, source_type: str = "close",
         sequential: bool = False) -> VossFilter:
    """
    Voss indicator by John F. Ehlers

    :param candles: np.ndarray
    :param period: int - default: 20
    :param predict: int - default: 3
    :param bandwith: float - default: 0.25
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """

    candles = slice_candles(candles, sequential)

    source = get_candle_source(candles, source_type=source_type)
    voss_val, filt = voss_fast(source, period, predict, bandwith)

    if sequential:
        return VossFilter(voss_val, filt)
    else:
        return VossFilter(voss_val[-1], filt[-1])


@njit(cache=True)
def voss_fast(source, period, predict, bandwith):
    voss = np.full_like(source, 0)
    filt = np.full_like(source, 0)

    pi = np.pi

    order = 3 * predict
    f1 = np.cos(2 * pi / period)
    g1 = np.cos(bandwith * 2 * pi / period)
    s1 = 1 / g1 - np.sqrt(1 / (g1 * g1) - 1)

    for i in range(source.shape[0]):
        if i > period and i > 5 and i > order:
            filt[i] = 0.5 * (1 - s1) * (source[i] - source[i - 2]) + f1 * (1 + s1) * filt[i - 1] - s1 * filt[i - 2]

    for i in range(source.shape[0]):
        if not (i <= period or i <= 5 or i <= order):
            sumc = 0
            for count in range(order):
                sumc = sumc + ((count + 1) / float(order)) * voss[i - (order - count)]
            voss[i] = ((3 + order) / 2) * filt[i] - sumc
    return voss, filt



from jesse.strategies import Strategy
import jesse.helpers as jh
from jesse import utils


class TestCanSubmitTakeProfitAndStopLossAtSameTimeInSpot(Strategy):
    def should_long(self):
        return self.price in [10, 20]

    def go_long(self):
        if self.price == 10:
            self.buy = 1, self.price
        elif self.price == 20:
            self.buy = 2, self.price

    def on_open_position(self, order) -> None:
        if self.price == 10:
            self.take_profit = self.position.qty, 15
            self.stop_loss = self.position.qty, 8
        elif self.price == 20:
            self.take_profit = self.position.qty, 30
            self.stop_loss = self.position.qty, 15

    def on_close_position(self, order) -> None:
        last_trade = self.trades[-1]

        if self.trades_count == 1:
            assert last_trade.exit_price == 15
            assert last_trade.entry_price == 10
            assert last_trade.qty == 1
            assert last_trade.type == "long"
        elif self.trades_count == 2:
            assert last_trade.exit_price == 30
            assert last_trade.entry_price == 20
            assert last_trade.qty == 2

    def should_cancel_entry(self):
        return False



from functools import lru_cache
from typing import Union

import numpy as np

import jesse.helpers as jh
import jesse.services.selectors as selectors
from jesse.enums import trade_types, order_types
from jesse.exceptions import EmptyPosition, OpenPositionError
from jesse.models import Order, Exchange
from jesse.services import logger
from jesse.utils import sum_floats, subtract_floats


class Position:
    def __init__(self, exchange_name: str, symbol: str, attributes: dict = None) -> None:
        self.id = jh.generate_unique_id()
        self.entry_price = None
        self.exit_price = None
        self.current_price = None
        self.qty = 0
        self.previous_qty = 0
        self.opened_at = None
        self.closed_at = None
        self._mark_price = None
        self._funding_rate = None
        self._next_funding_timestamp = None
        self._liquidation_price = None

        if attributes is None:
            attributes = {}

        self.exchange_name = exchange_name
        self.exchange: Exchange = selectors.get_exchange(self.exchange_name)

        self.symbol = symbol
        self.strategy = None

        for a in attributes:
            setattr(self, a, attributes[a])

    @property
    def mark_price(self) -> float:
        if not jh.is_live():
            return self.current_price

        if self.exchange_type == 'spot':
            return self.current_price

        return self._mark_price

    @property
    def funding_rate(self) -> float:
        if not jh.is_live():
            return 0

        if self.exchange_type == 'spot':
            raise ValueError('funding rate is not applicable to spot trading')

        return self._funding_rate

    @property
    def next_funding_timestamp(self) -> Union[int, None]:
        if not jh.is_live():
            return None

        if self.exchange_type == 'spot':
            raise ValueError('funding rate is not applicable to spot trading')

        return self._next_funding_timestamp

    @property
    def value(self) -> float:
        """
        The value of open position in the quote currency

        :return: float
        """
        if self.is_close:
            return 0

        if self.current_price is None:
            return None

        return abs(self.current_price * self.qty)

    @property
    def type(self) -> str:
        """
        The type of open position - long, short, or close

        :return: str
        """
        if self.is_long:
            return 'long'
        elif self.is_short:
            return 'short'

        return 'close'

    @property
    def pnl_percentage(self) -> float:
        """
        Alias for self.roi

        :return: float
        """
        return self.roi

    @property
    def roi(self) -> float:
        """
        Return on Investment in percentage
        More at: https://www.binance.com/en/support/faq/5b9ad93cb4854f5990b9fb97c03cfbeb
        """
        if self.pnl == 0:
            return 0

        return self.pnl / self.total_cost * 100

    @property
    def total_cost(self) -> float:
        """
        How much we paid to open this position (currently does not include fees, should we?!)
        """
        if self.is_close:
            return np.nan

        base_cost = self.entry_price * abs(self.qty)
        if self.strategy:
            return base_cost / self.leverage

        return base_cost

    @property
    def leverage(self) -> Union[int, np.float64]:
        if self.exchange_type == 'spot':
            return 1

        if self.strategy:
            return self.strategy.leverage
        else:
            return np.nan

    @property
    def exchange_type(self) -> str:
        return self.exchange.type

    @property
    def entry_margin(self) -> float:
        """
        Alias for self.total_cost
        """
        return self.total_cost

    @property
    def pnl(self) -> float:
        """
        The PNL of the position

        :return: float
        """
        if abs(self.qty) < self._min_qty:
            return 0

        if self.entry_price is None:
            return 0

        if self.value is None:
            return 0

        diff = self.value - abs(self.entry_price * self.qty)

        return -diff if self.type == 'short' else diff

    @property
    def is_open(self) -> bool:
        """
        Is the current position open?

        :return: bool
        """
        return self.type in ['long', 'short']

    @property
    def is_close(self) -> bool:
        """
        Is the current position close?

        :return: bool
        """
        return self.type == 'close'

    @property
    def is_long(self) -> bool:
        """
        Is the current position a long position?

        :return: bool
        """
        return self.qty > self._min_qty

    @property
    def is_short(self) -> bool:
        """
        Is the current position a short position?

        :return: bool
        """
        return self.qty < -abs(self._min_qty)

    @property
    def mode(self) -> str:
        if self.exchange.type == 'spot':
            return 'spot'
        else:
            return self.exchange.futures_leverage_mode

    @property
    def liquidation_price(self) -> Union[float, np.float64]:
        """
        The price at which the position gets liquidated. formulas are taken from:
        https://help.bybit.com/hc/en-us/articles/900000181046-Liquidation-Price-USDT-Contract-
        """
        if self.is_close:
            return np.nan

        if jh.is_livetrading():
            return self._liquidation_price

        if self.mode in ['cross', 'spot']:
            return np.nan

        elif self.mode == 'isolated':
            if self.type == 'long':
                return self.entry_price * (1 - self._initial_margin_rate + 0.004)
            elif self.type == 'short':
                return self.entry_price * (1 + self._initial_margin_rate - 0.004)
            else:
                return np.nan

        else:
            raise ValueError

    @property
    def _initial_margin_rate(self) -> float:
        return 1 / self.leverage

    @property
    def bankruptcy_price(self) -> Union[float, np.float64]:
        if self.type == 'long':
            return self.entry_price * (1 - self._initial_margin_rate)
        elif self.type == 'short':
            return self.entry_price * (1 + self._initial_margin_rate)
        else:
            return np.nan

    @property
    def to_dict(self):
        return {
            'entry_price': self.entry_price,
            'qty': self.qty,
            'current_price': self.current_price,
            'value': self.value,
            'type': self.type,
            'exchange': self.exchange_name,
            'pnl': self.pnl,
            'pnl_percentage': self.pnl_percentage,
            'leverage': self.leverage,
            'liquidation_price': self.liquidation_price,
            'bankruptcy_price': self.bankruptcy_price,
            'mode': self.mode,
        }

    def _mutating_close(self, close_price: float) -> None:
        if self.is_close and self._can_mutate_qty:
            raise EmptyPosition('The position is already closed.')

        self.exit_price = close_price
        self.closed_at = jh.now_to_timestamp()

        if self.exchange and self.exchange.type == 'futures':
            # just to prevent confusion
            close_qty = abs(self.qty)
            estimated_profit = jh.estimate_PNL(
                close_qty, self.entry_price,
                close_price, self.type
            )
            self.exchange.add_realized_pnl(estimated_profit)
            self.exchange.temp_reduced_amount[jh.base_asset(self.symbol)] += abs(close_qty * close_price)

        if self._can_mutate_qty:
            self._update_qty(0, operation='set')

        # reset entry_price
        self.entry_price = None

        self._close()

    def _close(self):
        from jesse.store import store
        store.completed_trades.close_trade(self)

    def _mutating_reduce(self, qty: float, price: float) -> None:
        if not self._can_mutate_qty:
            return

        if self.is_open is False:
            raise EmptyPosition('The position is closed.')

        # just to prevent confusion
        qty = abs(qty)

        estimated_profit = jh.estimate_PNL(qty, self.entry_price, price, self.type)

        if self.exchange and self.exchange.type == 'futures':
            # self.exchange.increase_futures_balance(qty * self.entry_price + estimated_profit)
            self.exchange.add_realized_pnl(estimated_profit)
            self.exchange.temp_reduced_amount[jh.base_asset(self.symbol)] += abs(qty * price)

        if self.type == trade_types.LONG:
            self._update_qty(qty, operation='subtract')
        elif self.type == trade_types.SHORT:
            self._update_qty(qty, operation='add')

    def _mutating_increase(self, qty: float, price: float) -> None:
        if not self.is_open:
            raise OpenPositionError('position must be already open in order to increase its size')

        qty = abs(qty)

        self.entry_price = jh.estimate_average_price(
            qty, price, self.qty,
            self.entry_price
        )

        if self._can_mutate_qty:
            if self.type == trade_types.LONG:
                self._update_qty(qty, operation='add')
            elif self.type == trade_types.SHORT:
                self._update_qty(qty, operation='subtract')

    def _mutating_open(self, qty: float, price: float) -> None:
        if self.is_open and self._can_mutate_qty:
            raise OpenPositionError('an already open position cannot be opened')

        self.entry_price = price
        self.exit_price = None

        if self._can_mutate_qty:
            self._update_qty(qty, operation='set')

        self.opened_at = jh.now_to_timestamp()

        self._open()

    def _update_qty(self, qty: float, operation='set'):
        self.previous_qty = self.qty

        if self.exchange_type == 'spot':
            if operation == 'set':
                self.qty = qty * (1 - self.exchange.fee_rate)
            elif operation == 'add':
                self.qty = sum_floats(self.qty, qty * (1 - self.exchange.fee_rate))
            elif operation == 'subtract':
                # fees are taken from the quote currency. in spot mode, sell orders cause
                # the qty to reduce but fees are handled on the exchange balance stuff
                self.qty = subtract_floats(self.qty, qty)

        elif self.exchange_type == 'futures':
            if operation == 'set':
                self.qty = qty
            elif operation == 'add':
                self.qty = sum_floats(self.qty, qty)
            elif operation == 'subtract':
                self.qty = subtract_floats(self.qty, qty)
        else:
            raise NotImplementedError('exchange type not implemented')

    def _open(self):
        from jesse.store import store
        store.completed_trades.open_trade(self)

    def _on_executed_order(self, order: Order) -> None:
        # futures (live)
        if jh.is_livetrading() and self.exchange_type == 'futures':
            # if position got closed because of this order
            if order.is_partially_filled:
                before_qty = self.qty - order.filled_qty
            else:
                before_qty = self.qty - order.qty
            after_qty = self.qty
            if before_qty != 0 and after_qty == 0:
                self._close()
        # spot (live)
        elif jh.is_livetrading() and self.exchange_type == 'spot':
            # if position got closed because of this order
            before_qty = self.previous_qty
            after_qty = self.qty
            qty = order.qty
            price = order.price
            closing_position = before_qty > self._min_qty > after_qty
            if closing_position:
                self._mutating_close(price)
            opening_position = before_qty < self._min_qty < after_qty
            if opening_position:
                self._mutating_open(qty, price)
            increasing_position = after_qty > before_qty > self._min_qty
            if increasing_position:
                self._mutating_increase(qty, price)
            reducing_position = self._min_qty < after_qty < before_qty
            if reducing_position:
                self._mutating_reduce(qty, price)
        else:  # backtest (both futures and spot)
            qty = order.qty
            price = order.price

            if self.exchange and self.exchange.type == 'futures':
                self.exchange.charge_fee(qty * price)

            # order opens position
            if self.qty == 0:
                change_balance = order.type == order_types.MARKET
                self._mutating_open(qty, price)
            # order closes position
            elif (sum_floats(self.qty, qty)) == 0:
                self._mutating_close(price)
            # order increases the size of the position
            elif self.qty * qty > 0:
                if order.reduce_only:
                    logger.info('Did not increase position because order is a reduce_only order')
                else:
                    self._mutating_increase(qty, price)
            # order reduces the size of the position
            elif self.qty * qty < 0:
                # if size of the order is big enough to both close the
                # position AND open it on the opposite side
                if abs(qty) > abs(self.qty):
                    if order.reduce_only:
                        logger.info(
                            f'Executed order is bigger than the current position size but it is a reduce_only order so it just closes it. Order QTY: {qty}, Position QTY: {self.qty}')
                        self._mutating_close(price)
                    else:
                        logger.info(
                            f'Executed order is big enough to not close, but flip the position type. Order QTY: {qty}, Position QTY: {self.qty}')
                        diff_qty = sum_floats(self.qty, qty)
                        self._mutating_close(price)
                        self._mutating_open(diff_qty, price)
                else:
                    self._mutating_reduce(qty, price)

        if self.strategy:
            self.strategy._on_updated_position(order)

    def update_from_stream(self, data: dict, is_initial: bool) -> None:
        """
        Used for updating the position from the WS stream (only for live trading)
        """
        before_qty = abs(self.qty)
        after_qty = abs(data['qty'])

        if self.exchange_type == 'futures':
            self.entry_price = data['entry_price']
            self._liquidation_price = data['liquidation_price']
        else:  # spot
            if after_qty > self._min_qty and self.entry_price is None:
                self.entry_price = self.current_price

        # if the new qty (data['qty']) is different than the current (self.qty) then update it:
        if self.qty != data['qty']:
            self.previous_qty = self.qty
            self.qty = data['qty']

        opening_position = before_qty <= self._min_qty < after_qty
        closing_position = before_qty > self._min_qty >= after_qty
        if opening_position:
            if is_initial:
                from jesse.store import store
                store.completed_trades.add_order_record_only(
                    self.exchange_name, self.symbol, jh.type_to_side(self.type),
                    self.qty, self.entry_price
                )
            self.opened_at = jh.now_to_timestamp()
            self._open()
        elif closing_position:
            self.closed_at = jh.now_to_timestamp()

    @property
    def _min_notional_size(self) -> float:
        if not (jh.is_livetrading() and self.exchange_type == 'spot'):
            return 0

        return self.exchange.vars['precisions'][self.symbol]['min_notional_size']

    @property
    @lru_cache
    def _min_qty(self) -> float:
        if not (jh.is_livetrading() and self.exchange_type == 'spot'):
            return 0

        # first check exchange return min_qty or not
        if 'min_qty' in self.exchange.vars['precisions'][self.symbol]:
            return self.exchange.vars['precisions'][self.symbol]['min_qty']

        if self._min_notional_size and self.current_price:
            return self._min_notional_size / self.current_price
        else:
            return 0

    @property
    def _can_mutate_qty(self):
        return not (self.exchange_type == 'spot' and jh.is_livetrading())



from collections import namedtuple

from jesse.indicators.roc import roc as _roc
from jesse.indicators.sma import sma as _sma
import numpy as np

from jesse.helpers import get_candle_source, slice_candles

KST = namedtuple('KST', ['line', 'signal'])


def kst(candles: np.ndarray, sma_period1: int = 10, sma_period2: int = 10, sma_period3: int = 10, sma_period4: int = 15,
        roc_period1: int = 10, roc_period2: int = 15, roc_period3: int = 20, roc_period4: int = 30,
        signal_period: int = 9, source_type: str = "close", sequential: bool = False) -> KST:
    """
    Know Sure Thing (KST)

    :param candles: np.ndarray
    :param sma_period1: int - default: 10
    :param sma_period2: int - default: 10
    :param sma_period3: int - default: 10
    :param sma_period4: int - default: 15
    :param roc_period1: int - default: 10
    :param roc_period2: int - default: 15
    :param roc_period3: int - default: 20
    :param roc_period4: int - default: 30
    :param signal_period: int - default: 9
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: KST(line, signal)
    """
    candles = slice_candles(candles, sequential)
    source = get_candle_source(candles, source_type=source_type)

    aroc1 = _sma(_roc(source, roc_period1, sequential=True), sma_period1, sequential=True)
    aroc2 = _sma(_roc(source, roc_period2, sequential=True), sma_period2, sequential=True)
    aroc3 = _sma(_roc(source, roc_period3, sequential=True), sma_period3, sequential=True)
    aroc4 = _sma(_roc(source, roc_period4, sequential=True), sma_period4, sequential=True)

    # Align arrays so that all have the same length as aroc4
    aligned_len = aroc4.size
    aroc1_aligned = aroc1[aroc1.size - aligned_len:]
    aroc2_aligned = aroc2[aroc2.size - aligned_len:]
    aroc3_aligned = aroc3[aroc3.size - aligned_len:]

    line = aroc1_aligned + 2 * aroc2_aligned + 3 * aroc3_aligned + 4 * aroc4
    signal = _sma(line, signal_period, sequential=True)

    if sequential:
        return KST(line, signal)
    else:
        return KST(line[-1], signal[-1])



from typing import Union

import numpy as np

from jesse.helpers import get_candle_source, slice_candles


def linearreg_angle(candles: np.ndarray, period: int = 14, source_type: str = "close", sequential: bool = False) -> \
        Union[float, np.ndarray]:
    """
    LINEARREG_ANGLE - Linear Regression Angle

    :param candles: np.ndarray
    :param period: int - default: 14
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    N = len(source)
    res = np.full(N, np.nan)
    if N >= period:
        # Create rolling windows of length 'period'
        windows = np.lib.stride_tricks.sliding_window_view(source, window_shape=period)
        x = np.arange(period)
        sum_x = x.sum()
        sum_x2 = (x * x).sum()
        common_den = period * sum_x2 - sum_x ** 2
        sum_y = np.sum(windows, axis=1)
        sum_xy = windows.dot(x)
        slopes = (period * sum_xy - sum_x * sum_y) / common_den
        angles = np.degrees(np.arctan(slopes))
        res[period - 1:] = angles
    return res if sequential else res[-1]



import jesse.helpers as jh
import jesse.services.logger as logger
from jesse.store import store


def set_up():
    store.reset()

# TODO
# def test_can_log_error_by_firing_event():
#     set_up()
#
#     # fire first error event
#     logger.error('first error!!!!!')
#     first_logged_error = {'id': 0, 'time': jh.now_to_timestamp(), 'message': 'first error!!!!!'}
#
#     assert store.logs.errors == [first_logged_error]
#
#     # fire second error event
#     logger.error('second error!!!!!')
#     second_logged_error = {'id': 1, 'time': jh.now_to_timestamp(), 'message': 'second error!!!!!'}
#
#     assert store.logs.errors == [first_logged_error, second_logged_error]
#
# TODO
# def test_can_log_info_by_firing_event():
#     set_up()
#
#     # fire first info event
#     logger.info('first info!!!!!')
#     first_logged_info = {'id': 0, 'time': jh.now_to_timestamp(), 'message': 'first info!!!!!'}
#
#     assert store.logs.info == [first_logged_info]
#
#     # fire second info event
#     logger.info('second info!!!!!')
#     second_logged_info = {
#         'id': 1,
#         'time': jh.now_to_timestamp(),
#         'message': 'second info!!!!!'
#     }
#
#     assert store.logs.info == [first_logged_info, second_logged_info]



from typing import Optional
from fastapi import APIRouter, Header
from fastapi.responses import JSONResponse

from jesse.services import auth as authenticator
from jesse.services.web import StoreNotificationApiKeyRequestJson, DeleteNotificationApiKeyRequestJson

router = APIRouter(prefix="/notification", tags=["Notification"])


@router.get("/api-keys")
def get_notification_api_keys(authorization: Optional[str] = Header(None)) -> JSONResponse:
    """
    Get all notification API keys
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.modes.notification_api_keys import get_notification_api_keys

    return get_notification_api_keys()


@router.post("/api-keys/store")
def store_notification_api_keys(
        json_request: StoreNotificationApiKeyRequestJson,
        authorization: Optional[str] = Header(None)
) -> JSONResponse:
    """
    Store a new notification API key
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.modes.notification_api_keys import store_notification_api_keys

    return store_notification_api_keys(
        json_request.name, json_request.driver, json_request.fields
    )


@router.post("/api-keys/delete")
def delete_notification_api_keys(
        json_request: DeleteNotificationApiKeyRequestJson,
        authorization: Optional[str] = Header(None)
) -> JSONResponse:
    """
    Delete a notification API key
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.modes.notification_api_keys import delete_notification_api_keys

    return delete_notification_api_keys(json_request.id)



from typing import Union

import numpy as np

from jesse.helpers import slice_candles


def correl(candles: np.ndarray, period: int = 5, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    CORREL - Pearson's Correlation Coefficient (r)

    :param candles: np.ndarray
    :param period: int - default: 5
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)
    x = candles[:, 3]
    y = candles[:, 4]
    n = len(x)
    
    # If not enough data, return an array of NaNs
    if n < period:
        res = np.empty(n)
        res[:] = np.nan
        return res if sequential else res[-1]

    # Use numpy's sliding_window_view for vectorized rolling window computation
    windows_x = np.lib.stride_tricks.sliding_window_view(x, window_shape=period)
    windows_y = np.lib.stride_tricks.sliding_window_view(y, window_shape=period)
    
    mean_x = np.mean(windows_x, axis=1)
    mean_y = np.mean(windows_y, axis=1)
    
    # Calculate numerator and denominator for Pearson correlation coefficient
    numerator = np.sum((windows_x - mean_x[:, None]) * (windows_y - mean_y[:, None]), axis=1)
    denominator = np.sqrt(np.sum((windows_x - mean_x[:, None])**2, axis=1) * np.sum((windows_y - mean_y[:, None])**2, axis=1))
    
    with np.errstate(divide='ignore', invalid='ignore'):
        corr_vals = numerator / denominator
    
    # Prepare full result array with initial NaNs for indices with insufficient data
    res_full = np.empty(n)
    res_full[:period-1] = np.nan
    res_full[period-1:] = corr_vals
    
    return res_full if sequential else res_full[-1]



import jesse.helpers as jh
from jesse.enums import sides
from jesse.exceptions import InsufficientBalance
from jesse.models import Order
from jesse.models.Exchange import Exchange
from jesse.enums import order_types
from jesse.utils import sum_floats, subtract_floats


class SpotExchange(Exchange):
    def __init__(self, name: str, starting_balance: float, fee_rate: float):
        super().__init__(name, starting_balance, fee_rate, 'spot')

        self.stop_orders_sum = {}
        self.limit_orders_sum = {}

        # # # # live-trading only # # # #
        self._started_balance = 0
        # # # # # # # # # # # # # # # # #

    @property
    def started_balance(self) -> float:
        if jh.is_livetrading():
            return self._started_balance

        return self.starting_assets[jh.app_currency()]

    @property
    def wallet_balance(self) -> float:
        return self.assets[self.settlement_currency]

    @property
    def available_margin(self) -> float:
        return self.wallet_balance

    def on_order_submission(self, order: Order) -> None:
        if jh.is_livetrading():
            return

        if order.side == sides.SELL:
            if order.type == order_types.STOP:
                self.stop_orders_sum[order.symbol] = sum_floats(self.stop_orders_sum.get(order.symbol, 0), abs(order.qty))
            elif order.type == order_types.LIMIT:
                self.limit_orders_sum[order.symbol] = sum_floats(self.limit_orders_sum.get(order.symbol, 0), abs(order.qty))

        base_asset = jh.base_asset(order.symbol)

        # buy order
        if order.side == sides.BUY:
            # cannot buy if we don't have enough balance (of the settlement currency)
            quote_balance = self.assets[self.settlement_currency]
            self.assets[self.settlement_currency] = subtract_floats(self.assets[self.settlement_currency], (abs(order.qty) * order.price))
            if self.assets[self.settlement_currency] < 0:
                raise InsufficientBalance(
                    f"Not enough balance. Available balance at {self.name} for {self.settlement_currency} is {quote_balance} but you're trying to spend {abs(order.qty * order.price)}"
                )
        # sell order
        else:
            base_balance = self.assets[base_asset]
            # sell order's qty cannot be bigger than the amount of existing base asset
            if order.type == order_types.MARKET:
                order_qty = sum_floats(abs(order.qty), self.limit_orders_sum.get(order.symbol, 0))
            elif order.type == order_types.STOP:
                order_qty = self.stop_orders_sum[order.symbol]
            elif order.type == order_types.LIMIT:
                order_qty = self.limit_orders_sum[order.symbol]
            else:
                raise Exception(f"Unknown order type {order.type}")
            # validate that the total selling amount is not bigger than the amount of the existing base asset
            if order_qty > base_balance:
                raise InsufficientBalance(
                    f"Not enough balance. Available balance at {self.name} for {base_asset} is {base_balance} but you're trying to sell {order_qty}"
                )

    def on_order_execution(self, order: Order) -> None:
        if jh.is_livetrading():
            return

        if order.side == sides.SELL:
            if order.type == order_types.STOP:
                self.stop_orders_sum[order.symbol] = subtract_floats(self.stop_orders_sum[order.symbol], abs(order.qty))
            elif order.type == order_types.LIMIT:
                self.limit_orders_sum[order.symbol] = subtract_floats(self.limit_orders_sum[order.symbol], abs(order.qty))

        base_asset = jh.base_asset(order.symbol)

        # buy order
        if order.side == sides.BUY:
            # asset's balance is increased by the amount of the order's qty after fees are deducted
            self.assets[base_asset] = sum_floats(self.assets[base_asset], abs(order.qty) * (1 - self.fee_rate))
        # sell order
        else:
            current_balance = self.assets[base_asset]
            if abs(order.qty) > current_balance:
                adjusted_qty = current_balance
                order_qty = abs(adjusted_qty)
            else:
                order_qty = abs(order.qty)

            # settlement currency's balance is increased by the amount of the order's qty after fees are deducted
            self.assets[self.settlement_currency] = sum_floats(
                self.assets[self.settlement_currency], (order_qty * order.price) * (1 - self.fee_rate)
            )
            # now reduce base asset's balance by the amount of the order's qty
            self.assets[base_asset] = subtract_floats(self.assets[base_asset], order_qty)

    def on_order_cancellation(self, order: Order) -> None:
        if jh.is_livetrading():
            return

        if order.side == sides.SELL:
            if order.type == order_types.STOP:
                self.stop_orders_sum[order.symbol] = subtract_floats(self.stop_orders_sum[order.symbol], abs(order.qty))
            elif order.type == order_types.LIMIT:
                self.limit_orders_sum[order.symbol] = subtract_floats(self.limit_orders_sum[order.symbol], abs(order.qty))

        base_asset = jh.base_asset(order.symbol)

        # buy order
        if order.side == sides.BUY:
            self.assets[self.settlement_currency] = sum_floats(self.assets[self.settlement_currency], abs(order.qty) * order.price)
        # sell order
        else:
            if order.type == order_types.STOP:
                self.stop_orders_sum[order.symbol] = subtract_floats(self.stop_orders_sum[order.symbol], abs(order.qty))
            elif order.type == order_types.LIMIT:
                self.limit_orders_sum[order.symbol] = subtract_floats(self.limit_orders_sum[order.symbol], abs(order.qty))


    def update_from_stream(self, data: dict) -> None:
        """
        Used for updating the exchange from the WS stream (only for live trading)
        """
        import jesse.services.logger as logger

        if not jh.is_livetrading():
            raise Exception('This method is only for live trading')

        old_balance = self.assets[self.settlement_currency]
        self.assets[self.settlement_currency] = data['balance']
        if old_balance != 0 and self.assets[self.settlement_currency] != old_balance:
            logger.info(
                f'Balance for {self.settlement_currency} on {self.name} changed from {round(old_balance, 2)} to {round(self.assets[self.settlement_currency], 2)}'
            )
        if self._started_balance == 0:
            self._started_balance = data['balance']



from jesse.strategies import Strategy


# test_shared_vars [part 1]
class Test32(Strategy):
    def __init__(self) -> None:
        super().__init__()

        self.shared_vars['buy-eth'] = False

    def before(self):
        if self.index == 10:
            self.shared_vars['buy-eth'] = True

    def should_long(self):
        return False

    def should_short(self):
        return False

    def go_long(self):
        pass

    def go_short(self):
        pass

    def should_cancel_entry(self):
        return False



from .acosc import acosc
from .ad import ad
from .adosc import adosc
from .adx import adx
from .adxr import adxr
from .alligator import alligator
from .alma import alma
from .ao import ao
from .apo import apo
from .aroon import aroon
from .aroonosc import aroonosc
from .atr import atr
from .avgprice import avgprice
from .bandpass import bandpass
from .beta import beta
from .bollinger_bands import bollinger_bands
from .bollinger_bands_width import bollinger_bands_width
from .bop import bop
from .cc import cc
from .cci import cci
from .cfo import cfo
from .cg import cg
from .chande import chande
from .chop import chop
from .cksp import cksp
from .cmo import cmo
from .correl import correl
from .correlation_cycle import correlation_cycle
from .cvi import cvi
from .cwma import cwma
from .damiani_volatmeter import damiani_volatmeter
from .dec_osc import dec_osc
from .decycler import decycler
from .dema import dema
from .devstop import devstop
from .di import di
from .dm import dm
from .donchian import donchian
from .dpo import dpo
from .dti import dti
from .dx import dx
from .edcf import edcf
from .efi import efi
from .ema import ema
from .emd import emd
from .emv import emv
from .epma import epma
from .er import er
from .eri import eri
from .fisher import fisher
from .fosc import fosc
from .frama import frama
from .fwma import fwma
from .gatorosc import gatorosc
from .gauss import gauss
from .heikin_ashi_candles import heikin_ashi_candles
from .high_pass import high_pass
from .high_pass_2_pole import high_pass_2_pole
from .hma import hma
from .hurst_exponent import hurst_exponent
from .hwma import hwma
from .ichimoku_cloud import ichimoku_cloud
from .ichimoku_cloud_seq import ichimoku_cloud_seq
from .ift_rsi import ift_rsi
from .itrend import itrend
from .jma import jma
from .jsa import jsa
from .kama import kama
from .kaufmanstop import kaufmanstop
from .kdj import kdj
from .keltner import keltner
from .kst import kst
from .kurtosis import kurtosis
from .kvo import kvo
from .linearreg import linearreg
from .linearreg_angle import linearreg_angle
from .linearreg_intercept import linearreg_intercept
from .linearreg_slope import linearreg_slope
from .lrsi import lrsi
from .ma import ma
from .maaq import maaq
from .mab import mab
from .macd import macd
from .mama import mama
from .marketfi import marketfi
from .mass import mass
from .mcginley_dynamic import mcginley_dynamic
from .mean_ad import mean_ad
from .median_ad import median_ad
from .medprice import medprice
from .mfi import mfi
from .midpoint import midpoint
from .midprice import midprice
from .minmax import minmax
from .mom import mom
from .mwdx import mwdx
from .natr import natr
from .nma import nma
from .nvi import nvi
from .obv import obv
from .pfe import pfe
from .pivot import pivot
from .pma import pma
from .ppo import ppo
from .pvi import pvi
from .pwma import pwma
from .qstick import qstick
from .reflex import reflex
from .rma import rma
from .roc import roc
from .rocp import rocp
from .rocr import rocr
from .rocr100 import rocr100
from .roofing import roofing
from .rsi import rsi
from .rsmk import rsmk
from .rsx import rsx
from .rvi import rvi
from .safezonestop import safezonestop
from .sar import sar
from .sinwma import sinwma
from .skew import skew
from .sma import sma
from .smma import smma
from .sqwma import sqwma
from .srsi import srsi
from .srwma import srwma
from .stc import stc
from .stddev import stddev
from .stochastic import stoch
from .stochf import stochf
from .supersmoother import supersmoother
from .supersmoother_3_pole import supersmoother_3_pole
from .supertrend import supertrend
from .swma import swma
from .t3 import t3
from .tema import tema
from .trange import trange
from .trendflex import trendflex
from .trima import trima
from .trix import trix
from .tsf import tsf
from .tsi import tsi
from .ttm_trend import ttm_trend
from .typprice import typprice
from .ui import ui
from .ultosc import ultosc
from .var import var
from .vi import vi
from .vidya import vidya
from .vlma import vlma
from .vosc import vosc
from .voss import voss
from .vpci import vpci
from .vpt import vpt
from .vpwma import vpwma
from .vwap import vwap
from .vwma import vwma
from .vwmacd import vwmacd
from .wad import wad
from .wclprice import wclprice
from .wilders import wilders
from .willr import willr
from .wma import wma
from .wt import wt
from .zlema import zlema
from .zscore import zscore
from .waddah_attr_explosion import waddah_attar_explosion
from .stiffness import stiffness
from .ttm_squeeze import ttm_squeeze
from .support_resistance_with_break import support_resistance_with_breaks
from .squeeze_momentum import squeeze_momentum
from .hull_suit import hull_suit
from .volume import volume



from jesse.strategies import Strategy


class TestPositions(Strategy):
    def before(self):
        if self.index == 0:
            assert self.all_positions[self.symbol] == self.position

    def should_long(self) -> bool:
        return False

    def should_short(self) -> bool:
        return False

    def go_long(self):
        pass

    def go_short(self):
        pass

    def should_cancel_entry(self):
        return False



from jesse import utils
from jesse.strategies import Strategy
import jesse.helpers as jh


class TestWalletBalance(Strategy):
    def before(self):
        if self.index == 0:
            assert self.position.exchange.wallet_balance == 10000 == self.balance
            assert self.position.exchange.available_margin == 10000 == self.available_margin
            assert self.leveraged_available_margin == 20000

        if self.price == 11:
            # wallet balance should have stayed the same while we haven't spent from it yet
            assert self.position.exchange.wallet_balance == 10000 == self.balance
            # Adjusted available_margin calculation
            assert round(self.position.exchange.available_margin) == 10000 - (4000 / 2) == round(self.available_margin)
            assert round(self.leveraged_available_margin) == 20000 - 4000

        if self.price == 12:
            # wallet balance should have changed because of fees, but we have fee=0 in this test, so:
            assert self.position.exchange.wallet_balance == 10000 == self.balance
            # Adjusted available_margin calculation
            assert round(self.position.exchange.available_margin) == 10000 - (4000 / 2) == round(self.available_margin)

        if self.price == 21:
            # wallet balance must now equal to 10_000 + the profit we made from previous trade
            previous_trade = self.trades[0]
            assert self.position.exchange.wallet_balance == previous_trade.pnl + 10000
            # now that position is closed, available_margin should equal to wallet_balance
            assert self.position.exchange.available_margin == previous_trade.pnl + 10000 == self.available_margin
            assert self.balance == self.position.exchange.wallet_balance
            assert self.balance == self.available_margin

    def should_long(self) -> bool:
        return self.price == 10

    def should_short(self) -> bool:
        return False

    def go_long(self):
        qty1 = utils.size_to_qty(2000, self.price + 2)
        qty2 = utils.size_to_qty(2000, self.price + 4)
        self.buy = [
            (qty1, self.price + 2),
            (qty2, self.price + 4),
        ]

    def update_position(self):
        if self.price == 20:
            self.liquidate()

    def go_short(self):
        pass

    def should_cancel_entry(self):
        return False



import jesse.helpers as jh
from jesse.config import reset_config
from jesse.enums import exchanges
from jesse.factories import candles_from_close_prices
from jesse.modes import backtest_mode
from jesse.config import config


def get_btc_and_eth_candles():
    candles = {
        jh.key(exchanges.SANDBOX, 'BTC-USDT'): {
            'exchange': exchanges.SANDBOX,
            'symbol': 'BTC-USDT',
            'candles': candles_from_close_prices(range(101, 200)),
        }
    }

    candles[jh.key(exchanges.SANDBOX, 'ETH-USDT')] = {
        'exchange': exchanges.SANDBOX,
        'symbol': 'ETH-USDT',
        'candles': candles_from_close_prices(range(1, 100))
    }
    return candles


def get_btc_candles(candles_count=100):
    return {
        jh.key(exchanges.SANDBOX, 'BTC-USDT'): {
            'exchange': exchanges.SANDBOX,
            'symbol': 'BTC-USDT',
            'candles': candles_from_close_prices(range(1, candles_count)),
        }
    }


def get_downtrend_candles(candles_count=100):
    return {
        jh.key(exchanges.SANDBOX, 'BTC-USDT'): {
            'exchange': exchanges.SANDBOX,
            'symbol': 'BTC-USDT',
            'candles': candles_from_close_prices(range(candles_count, 10, -1)),
        }
    }


def set_up(is_futures_trading=True, leverage=1, leverage_mode='cross', fee=0):
    reset_config()
    config['env']['exchanges'][exchanges.SANDBOX]['balance'] = 10_000

    config['env']['exchanges']['Sandbox']['fee'] = fee

    if is_futures_trading:
        # used only in futures trading
        config['env']['exchanges'][exchanges.SANDBOX]['type'] = 'futures'
        config['env']['exchanges'][exchanges.SANDBOX]['futures_leverage_mode'] = leverage_mode
        config['env']['exchanges'][exchanges.SANDBOX]['futures_leverage'] = leverage
    else:
        config['env']['exchanges'][exchanges.SANDBOX]['type'] = 'spot'


def single_route_backtest(
        strategy_name: str, is_futures_trading=True,
        leverage=1, leverage_mode='cross', trend='up', fee=0,
        candles_count=100, timeframe='1m'
):
    """
    used to simplify simple tests
    """
    set_up(
        is_futures_trading=is_futures_trading,
        leverage=leverage,
        leverage_mode=leverage_mode,
        fee=fee
    )

    routes = [{'symbol': 'BTC-USDT', 'timeframe': timeframe, 'strategy': strategy_name}]

    if trend == 'up':
        candles = get_btc_candles(candles_count)
    elif trend == 'down':
        candles = get_downtrend_candles(candles_count)
    else:
        raise ValueError

    # dates are fake. just to pass required parameters
    backtest_mode.run('000', False, {}, exchanges.SANDBOX, routes, [], '2019-04-01', '2019-04-02', candles)


def two_routes_backtest(
        strategy_name1: str, strategy_name2: str, is_futures_trading=True, leverage=1, leverage_mode='cross', trend='up'
):
    """
    used to simplify simple tests
    """
    set_up(
        is_futures_trading=is_futures_trading,
        leverage=leverage,
        leverage_mode=leverage_mode
    )

    routes = [
        {'symbol': 'BTC-USDT', 'timeframe': '1m', 'strategy': strategy_name1},
        {'symbol': 'ETH-USDT', 'timeframe': '1m', 'strategy': strategy_name2},
    ]

    # dates are fake. just to pass required parameters
    backtest_mode.run('000', False, {}, exchanges.SANDBOX, routes, [], '2019-04-01', '2019-04-02', get_btc_and_eth_candles())


def two_data_routes_backtest(
        strategy_name1: str, strategy_name2: str, is_futures_trading=True, leverage=1, leverage_mode='cross', trend='up'
):
    """
    used to simplify simple tests
    """
    set_up(
        is_futures_trading=is_futures_trading,
        leverage=leverage,
        leverage_mode=leverage_mode
    )

    routes = [
        {'symbol': 'BTC-USDT', 'timeframe': '1m', 'strategy': strategy_name1},
        {'symbol': 'ETH-USDT', 'timeframe': '5m', 'strategy': strategy_name2},
    ]

    data_routes = [
        {'symbol': 'BTC-USDT', 'timeframe': '5m'},
        {'symbol': 'ETH-USDT', 'timeframe': '15m'},
    ]

    # dates are fake. just to pass required parameters
    backtest_mode.run('000', False, {}, exchanges.SANDBOX, routes, data_routes, '2019-04-01', '2019-04-02', get_btc_and_eth_candles())



from collections import namedtuple
from .macd import macd
from .sma import sma
from .stddev import stddev
import numpy as np
from jesse.helpers import get_candle_source, slice_candles

WaddahAttarExplosionTuple = namedtuple(
    'WaddahAttarExplosionTuple', ['explosion_line', 'trend_power', 'trend_direction']
)


def waddah_attar_explosion(candles: np.ndarray, sensitivity: int = 150, fast_length: int = 20, slow_length: int = 40, channel_length: int = 20, mult: float = 2.0, source_type: str = "close") -> WaddahAttarExplosionTuple:
    """
    @author LazyBear 
    credits: https://www.tradingview.com/v/iu3kKWDI/

    WADDAH_ATTAR_EXPLOSION - Waddah Attar Explosion

    :param candles: np.ndarray
    :param sensitivity: int - default: 150
    :param fast_length: int - default: 20
    :param slow_length: int - default: 40
    :param channel_length: int - default: 20
    :param mult: float - default: 2.0
    :param source_type: str - default: "close"

    :return: WaddahAttarExplosionTuple(explosion_line, trend_power, trend_direction)
    """
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, False)
        source = get_candle_source(candles, source_type=source_type)

    t1 = (macd(source, fast_period=fast_length, slow_period=slow_length)[0] -
          macd(source[:-1], fast_period=fast_length, slow_period=slow_length)[0])*sensitivity
    trend = 1 if t1 >= 0 else -1
    e1 = _calc_bb_upper(source, channel_length, mult) - _calc_bb_lower(source, channel_length, mult)

    return WaddahAttarExplosionTuple(e1, t1, trend)


def _calc_bb_upper(source, length, mult):
    basis = sma(source, length)
    dev = mult * stddev(source, length)
    return basis + dev


def _calc_bb_lower(source, length, mult):
    basis = sma(source, length)
    dev = mult * stddev(source, length)
    return basis - dev



from jesse.strategies import Strategy


# test_on_route_close_position part 2 - ETH-USD
class Test26(Strategy):
    def should_long(self):
        return False

    def should_short(self):
        return self.price == 10

    def go_long(self):
        pass

    def go_short(self):
        self.sell = 1, self.price
        self.stop_loss = 1, 20

    def should_cancel_entry(self):
        return False



from typing import Union

import numpy as np
from numpy.lib.stride_tricks import sliding_window_view

from jesse.helpers import get_candle_source, same_length, slice_candles


def mean_ad(candles: np.ndarray, period: int = 5, source_type: str = "hl2", sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    Mean Absolute Deviation

    :param candles: np.ndarray
    :param period: int - default: 5
    :param source_type: str - default: "hl2"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    if len(candles.shape) == 1:
      source = candles
    else:
      candles = slice_candles(candles, sequential)
      source = get_candle_source(candles, source_type=source_type)

    swv = sliding_window_view(source, window_shape=period)
    abs_diff = np.absolute(source - same_length(source, np.mean(swv, -1)))
    smv_abs_diff = sliding_window_view(abs_diff, window_shape=period)
    mean_abs_deviation = np.nanmean(smv_abs_diff, -1)
    res = same_length(source, mean_abs_deviation)

    return res if sequential else res[-1]




from jesse.services import logger
from jesse.strategies import Strategy


# test_end
class Test41(Strategy):
    def should_long(self) -> bool:
        return self.index == 0

    def should_short(self) -> bool:
        return False

    def go_long(self):
        qty = 1
        self.buy = qty, 2

    def go_short(self):
        pass

    def should_cancel_entry(self):
        return False

    def before_terminate(self):
        # log, so we can check this block was executed in the first place
        logger.info('executed terminate successfully')

        # assert open position
        assert self.position.is_open
        assert self.position.pnl == 97

        # close it manually
        self.liquidate()



from collections import namedtuple

from typing import Union

import numpy as np
from jesse.helpers import slice_candles
from jesse.indicators.rma import rma
from numba import njit

DX = namedtuple('DX', ['adx', 'plusDI', 'minusDI'])

@njit(cache=True)
def _fast_dm_tr(high: np.ndarray, low: np.ndarray, close: np.ndarray) -> tuple:
    n = len(high)
    up = np.zeros(n)
    down = np.zeros(n)
    plusDM = np.zeros(n)
    minusDM = np.zeros(n)
    true_range = np.zeros(n)
    
    for i in range(n):
        if i == 0:
            up[i] = 0
            down[i] = 0
            plusDM[i] = 0
            minusDM[i] = 0
            true_range[i] = high[i] - low[i]
        else:
            up[i] = high[i] - high[i - 1]
            down[i] = low[i - 1] - low[i]
            plusDM[i] = up[i] if (up[i] > down[i] and up[i] > 0) else 0
            minusDM[i] = down[i] if (down[i] > up[i] and down[i] > 0) else 0
            a = high[i] - low[i]
            b = abs(high[i] - close[i - 1])
            c = abs(low[i] - close[i - 1])
            true_range[i] = max(a, b, c)
    
    return plusDM, minusDM, true_range

def dx(candles: np.ndarray, di_length: int = 14, adx_smoothing: int = 14, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    DX - Directional Movement Index
    
    :param candles: np.ndarray
    :param di_length: int - default: 14
    :param adx_smoothing: int - default: 14
    :param sequential: bool - default: False
    
    :return: DX(adx, plusDI, minusDI)
    """
    candles = slice_candles(candles, sequential)
    high = candles[:, 3]
    low = candles[:, 4]
    close = candles[:, 2]
    
    plusDM, minusDM, true_range = _fast_dm_tr(high, low, close)
    
    tr_rma = rma(true_range, di_length, sequential=True)
    plus_rma = rma(plusDM, di_length, sequential=True)
    minus_rma = rma(minusDM, di_length, sequential=True)
    
    # Compute +DI and -DI, avoiding division by zero
    plusDI = np.where(tr_rma == 0, 0, 100 * plus_rma / tr_rma)
    minusDI = np.where(tr_rma == 0, 0, 100 * minus_rma / tr_rma)
    
    di_sum = plusDI + minusDI
    di_diff = np.abs(plusDI - minusDI)
    directional_index = di_diff / np.where(di_sum == 0, 1, di_sum)
    adx = 100 * rma(directional_index, adx_smoothing, sequential=True)
    
    if sequential:
        return DX(adx, plusDI, minusDI)
    else:
        return DX(adx[-1], plusDI[-1], minusDI[-1])



from typing import Union

import numpy as np
from numba import njit

from jesse.helpers import get_candle_source, slice_candles


@njit
def _ema_numba(data, period):
    N = len(data)
    result = np.empty(N, dtype=np.float64)
    alpha = 2.0 / (period + 1)
    result[0] = data[0]
    for i in range(1, N):
        result[i] = alpha * data[i] + (1 - alpha) * result[i - 1]
    return result


def ema(data: np.ndarray, period: int) -> np.ndarray:
    return _ema_numba(data, period)


def trix(candles: np.ndarray, period: int = 18, source_type: str = "close", sequential: bool = False) -> Union[float, np.ndarray]:
    """
    TRIX - 1-day Rate-Of-Change (ROC) of a Triple Smooth EMA

    :param candles: np.ndarray
    :param period: int - default: 18
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)
    source = get_candle_source(candles, source_type=source_type)

    # Compute triple EMA on the logarithm of the source prices
    log_source = np.log(source)
    ema1 = ema(log_source, period)
    ema2 = ema(ema1, period)
    ema3 = ema(ema2, period)

    # Calculate the change (current ema3 minus previous ema3), prepending NaN to maintain same length
    diff = np.empty_like(ema3)
    diff[0] = np.nan
    diff[1:] = ema3[1:] - ema3[:-1]
    result = diff * 10000

    return result if sequential else result[-1]



from typing import Union

import numpy as np

from jesse.helpers import get_candle_source, slice_candles


def rocr100(candles: np.ndarray, period: int = 10, source_type: str = "close", sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    ROCR100 - Rate of change ratio 100 scale: (price/prevPrice)*100

    :param candles: np.ndarray
    :param period: int - default: 10
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    source = get_candle_source(candles, source_type=source_type)
    # Vectorized calculation: for indices >= period, ROCR100 = (source[i] / source[i - period]) * 100; first period set to np.nan
    res = np.full(source.shape, np.nan, dtype=float)
    res[period:] = (source[period:] / source[:-period]) * 100
    return res if sequential else res[-1]



import sys
from math import log10
import jesse.helpers as jh
from jesse.research.backtest import _isolated_backtest as isolated_backtest
from jesse.services import logger
import numpy as np
from jesse import exceptions


def _formatted_inputs_for_isolated_backtest(user_config, routes):
    # Format input parameters required for backtest simulation
    return {
        'starting_balance': user_config['exchange']['balance'],
        'fee': user_config['exchange']['fee'],
        'type': user_config['exchange']['type'],
        'futures_leverage': user_config['exchange']['futures_leverage'],
        'futures_leverage_mode': user_config['exchange']['futures_leverage_mode'],
        'exchange': routes[0]['exchange'],
        'warm_up_candles': jh.get_config('env.data.warmup_candles_num')
    }


def get_fitness(
        user_config: dict, routes: list, data_routes: list, strategy_hp, hp: dict,
        training_warmup_candles: dict, training_candles: dict,
        testing_warmup_candles: dict, testing_candles: dict, optimal_total: int, fast_mode: bool
) -> tuple:
    """
    Evaluates the fitness (i.e. backtest performance) of the strategy
    using the given hyperparameters (hp). The fitness score is calculated based on the backtest results.
    """
    try:
        inputs = _formatted_inputs_for_isolated_backtest(user_config, routes)
        # Run backtest simulation for the training data using the suggested hyperparameters
        training_metrics = isolated_backtest(
            inputs,
            routes,
            data_routes,
            candles=training_candles,
            warmup_candles=training_warmup_candles,
            hyperparameters=hp,
            fast_mode=fast_mode
        )['metrics']

        # Calculate fitness score
        if training_metrics['total'] > 5:
            total_effect_rate = log10(training_metrics['total']) / log10(optimal_total)
            total_effect_rate = min(total_effect_rate, 1)
            objective_function_config = jh.get_config('env.optimization.objective_function', 'sharpe')
            
            # Get the ratio based on objective function
            if objective_function_config == 'sharpe':
                ratio = training_metrics['sharpe_ratio']
                ratio_normalized = jh.normalize(ratio, -.5, 5)
            elif objective_function_config == 'calmar':
                ratio = training_metrics['calmar_ratio']
                ratio_normalized = jh.normalize(ratio, -.5, 30)
            elif objective_function_config == 'sortino':
                ratio = training_metrics['sortino_ratio']
                ratio_normalized = jh.normalize(ratio, -.5, 15)
            elif objective_function_config == 'omega':
                ratio = training_metrics['omega_ratio']
                ratio_normalized = jh.normalize(ratio, -.5, 5)
            elif objective_function_config == 'serenity':
                ratio = training_metrics['serenity_index']
                ratio_normalized = jh.normalize(ratio, -.5, 15)
            elif objective_function_config == 'smart sharpe':
                ratio = training_metrics['smart_sharpe']
                ratio_normalized = jh.normalize(ratio, -.5, 5)
            elif objective_function_config == 'smart sortino':
                ratio = training_metrics['smart_sortino']
                ratio_normalized = jh.normalize(ratio, -.5, 15)
            else:
                raise ValueError(
                    f'The entered ratio configuration `{objective_function_config}` for the optimization is unknown. '
                    f'Choose between sharpe, calmar, sortino, serenity, smart sharpe, smart sortino and omega.'
                )

            # If the ratio is negative then the configuration is not usable
            if ratio < 0:
                score = 0.0001
                logger.log_optimize_mode(f"NEGATIVE RATIO: hp is not usable => {objective_function_config}: {ratio}, total: {training_metrics['total']}")
                return score, training_metrics, {}

            # Run backtest for testing period
            testing_metrics = isolated_backtest(
                inputs,
                routes,
                data_routes,
                candles=testing_candles,
                warmup_candles=testing_warmup_candles,
                hyperparameters=hp,
                fast_mode=fast_mode
            )['metrics']

            # Calculate fitness score
            score = total_effect_rate * ratio_normalized
            if np.isnan(score):
                logger.log_optimize_mode(f'Score is nan. hp configuration is invalid')
                score = 0.0001
            else:
                logger.log_optimize_mode(f"hp config is usable => {objective_function_config}: {round(ratio, 2)}, total: {training_metrics['total']}, "
                                       f"pnl%: {round(training_metrics['net_profit_percentage'], 2)}%, win-rate: {round(training_metrics['win_rate']*100, 2)}%")
        else:
            logger.log_optimize_mode('Less than 5 trades in the training data. hp configuration is invalid')
            score = 0.0001
            training_metrics = {}
            testing_metrics = {}

        return score, training_metrics, testing_metrics

    except exceptions.RouteNotFound as e:
        raise e
    except Exception as e:
        import sys, traceback
        exc_type, exc_value, exc_traceback = sys.exc_info()
        traceback_details = {
            "filename": exc_traceback.tb_frame.f_code.co_filename,
            "line": exc_traceback.tb_lineno,
            "name": exc_traceback.tb_frame.f_code.co_name,
            "type": exc_type.__name__,
            "message": str(e)
        }
        logger.log_optimize_mode(f"Trial evaluation failed: {traceback_details}")
        return 0.0001, {}, {}
